from .map_node import MapNode
from .vector_node import VectorNode

# Other nodes
from .dummy_identity_map_node import DummyIdentityMapNode
from .null_node import NullNode

# Custom nodes go here
from .rgb import RGB
from .hsv import HSV
from .depth import Depth
from .edges import Edges
from .normal import Normal
from .cameranormal import CameraNormal
from .semantic import Semantic
from .halftone import Halftone
from .wireframe import Wireframe
from .softsegmentation import SoftSegmentation
from .pose import Pose