from .ngc_edge import NGCEdge
from .single_link import SingleLink
from .two_hops_nogt import TwoHopsNoGT
from .single_link_residual import SingleLinkResidual
from .two_hops_residual_nogt import TwoHopsResidualNoGT
